/*
  appliance.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "appliance.h"

#include "dmp.h"
#include "sdt.h"
#include "rlp.h"
#include "rlp-udp.h"
#include "rlp-tcp.h"

namespace ACN {

/**
 * @brief Appliance::Appliance
 * @param cid
 */
Appliance::Appliance(UUID::uuid cid)
  : Component(cid)
{
  fctn_ = "libESTA ACN Appliance";

  /**
   * 1.3.1. Recursive use of PDU Format in Other Protocols
   * ... within ACN as currently specified, there is just one higher layer
   * protocol defined which interfaces directly to the root layer:
   *    Session Data Transport
   */
  RlpRegisterVector(SDT::SDT_PROTOCOL_ID, std::bind(&Appliance::SdtReceiver,
                                                  this, std::placeholders::_1));
};


/**
 * @brief Appliance::UdpPayloadReceiver
 * @param stream
 *
 * EPI 17
 */
void Appliance::UdpPayloadReceiver(PDU::Stream stream)
{
  auto transport = RLP::UDP::transport();
  transport.iStream(stream);

  for(auto const &root : *transport.root.pdu) {
    RlpReceiver(root);
  }
}


/**
 * @brief Appliance::TcpPacketReceiver
 * @param stream
 *
 * EPI 33
 */
void Appliance::TcpPacketReceiver(PDU::Stream stream)
{
  auto transport = RLP::TCP::transport();
  transport.iStream(stream);

  for(auto const &root : *transport.root.pdu) {
    RlpReceiver(root);
  }
}


/**
 * @brief Appliance::RlpReceiver
 * @param root
 *
 * Dispatch a recieved RLP PDU to the appropriate vector handlers.
 */
void Appliance::RlpReceiver(std::shared_ptr<RLP::Pdu> root)
{
  /// To this point, the data segment of the RLP PDU will not have been read.
  /// The root PDU must have not failed, and still have bytes available to read.
  if (!root->stream()->good())
    return;

  if (rlp_vectors_.count(root->vector()))
    for(auto const &handler : rlp_vectors_[root->vector()])
      handler(root);
}


/**
 * @brief Appliance::RlpRegisterVectorHandler
 * @param vect
 * @param handle
 *
 * Add callback handler for a given RLP vector.
 */
void Appliance::RlpRegisterVector(uint32_t vect, PDU::Handler<RLP::Pdu> handle)
{
  rlp_vectors_[vect].push_back(handle);
}


/**
 * @brief Appliance::SdtReceiver
 * @param rlp
 */
void Appliance::SdtReceiver(std::shared_ptr<RLP::Pdu> root)
{
  root->createDataBlock<SDT::Pdu>();
  auto block = static_cast<PDU::Block<SDT::Pdu>*>(root->data());

  for(auto const &sdt : *block->pdu) {
    switch(sdt->vector()) {
      /// TODO: handle SDT
      default:
        break;
    }
  }
}


/**
 * @brief Appliance::DmpReceiver
 * @param rlp
 */
void Appliance::DmpReceiver(std::shared_ptr<RLP::Pdu> root)
{
  root->createDataBlock<DMP::Pdu>();
  auto block = static_cast<PDU::Block<DMP::Pdu>*>(root->data());

  for(auto const &dmp : *block->pdu) {
    switch(dmp->vector()) {
      /// TODO: DMP in root
      default:
        break;
    }
  }
}

}; // ACN
