/*
  extended.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "acn/pdu.h"
#include "sacn.h"
#include "uuid/uuid.h"
#include <functional>

namespace SACN {
namespace EXTENDED {

/**
 * @brief 6.3 E1.31 Synchronization Packet Framing Layer
 */
struct sync_header
    : ACN::PDU::pdu_header
{
  uint8_t  sequence_number;
  uint16_t sync_address;
  uint16_t reserved;

  size_t streamSize() const override { return 5; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief 6.4 E1.31 Universe Discovery Packet Framing Layer
 */
struct discovery_header
    : ACN::PDU::pdu_header
{
  std::string source_name;
  uint32_t reserved;

  size_t streamSize() const override { return 68; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief The EXTENDED::Pdu class
 */
class Pdu
    : public ACN::PDU::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


namespace DISCOVERY {

/**
 * @brief Table 8-9: E1.31 Universe Discovery Packet Universe Discovery Layer
 */
struct discovery_list_header
    : ACN::PDU::pdu_header
{
  uint8_t page;
  uint8_t last_page;

  size_t streamSize() const override { return 2; }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief The discoveredUniverse struct
 */
struct discoveredUniverse
{
  UUID::uuid  source;
  std::string description;
  uint16_t    universe;
};

using Watcher = std::function<void(std::shared_ptr<discoveredUniverse>)>;


/**
 * @brief The discovery_list_data struct
 */
struct discovery_list_data
    : ACN::PDU::pdu_data
{
  std::vector<discoveredUniverse> found;

  size_t streamSize() const override { return 2 * found.size(); }
  void iStream(ACN::PDU::Stream) override;
  void oStream(ACN::PDU::Stream) const override;
};


/**
 * @brief The EXTENDED::DISCOVERY::Pdu class
 */
class Pdu
    : public ACN::PDU::Pdu
{
public:
  Pdu();
  void iStream(ACN::PDU::Stream) override;
};


} // DISCOVERY
} // EXTENDED
} // SACN
